import collections.abc

import toolz


def apply_middleware(*middlewares):
    def next_func(next_):
        # next_ is typically aioredux.create_store
        def create_store(reducer, initial_state=None):
            store = yield from next_(reducer, initial_state)
            dispatch = store.dispatch
            middleware_api = dict(dispatch=dispatch, state_func=lambda: store.state)
            chain = map(lambda middleware: middleware(**middleware_api), middlewares)
            store.dispatch = toolz.compose(*chain)(dispatch)
            return store
        return create_store
    return next_func


def is_mapping(action):
    '''Analogous to isPlainObject'''
    return isinstance(action, collections.abc.Mapping)


def is_FSA(action):
    '''Checks whether `action` is a Flux Standard Action (FSA).'''
    if not is_mapping(action):
        return False
    if 'type' not in action:
        return False
    properties = {'type', 'action', 'payload', 'error', 'meta'}
    if len(action.keys() - properties) > 0:
        return False
    if not is_mapping(action.get('payload', {})):
        return False
    return True


def combine_reducers(reducers):
    assert all(callable(val) for val in reducers.values())
    default_state = {k: None for k in reducers}

    def combination(state=None, action=None):
        if state is None:
            state = default_state
        has_changed = False

        def next_state(key, reducer):
            nonlocal has_changed
            previous_state_for_key = state[key]
            next_state_for_key = reducer(previous_state_for_key, action)
            if next_state_for_key is None:
                msg = ('`None` is not an allowed initial state when using `combine_reducers`. '
                       '`None` is used to indicate a malfunctioning reducer initialization.')
                raise ValueError(msg)
            has_changed |= next_state_for_key != previous_state_for_key
            return next_state_for_key
        final_state = {key: next_state(key, reducer) for key, reducer in reducers.items()}
        return final_state if has_changed else state
    return combination



import asyncio

import toolz

import aioredux
from aioredux.tests import base


class TestAioredux(base.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        super().setUp()

    def tearDown(self):
        self.loop.close()
        self.loop = None
        super().tearDown()

    def test_todo(self):

        @asyncio.coroutine
        def go():
            initial_state = {
                'todos': (),
            }
            store = yield from aioredux.create_store(lambda state, action: state, initial_state, loop=self.loop)
            self.assertIsNotNone(store)
            self.assertIsNotNone(store.state)

        self.loop.run_until_complete(go())

    def test_todo_reducer(self):

        @asyncio.coroutine
        def go():
            initial_state = {
                'todos': (),
            }

            def add_todo(text):
                return {'type': 'ADD_TODO', 'text': text}

            def reducer(state, action):
                if action['type'] == 'ADD_TODO':
                    todos = state['todos'] + (action['text'],)
                    return toolz.assoc(state, 'todos', todos)
                return state

            store = yield from aioredux.create_store(reducer, initial_state, loop=self.loop)
            self.assertIsNotNone(store.state)
            self.assertIsNotNone(store.state['todos'])
            self.assertEqual(len(store.state['todos']), 0)
            yield from store.dispatch(add_todo('todo text'))
            self.assertIsNotNone(store.state)
            self.assertIsNotNone(store.state['todos'])
            self.assertEqual(len(store.state['todos']), 1)

        self.loop.run_until_complete(go())



import asyncio
import enum
import logging

import toolz

import aioredux

logger = logging.getLogger(__name__)


# action types
class ActionTypes(enum.Enum):
    ADD_TODO = 1
    REMOVE_TODO = 2
    COMPLETE_TODO = 3


# action creators
def add_todo(text):
    return {'type': ActionTypes.ADD_TODO, 'text': text}


def complete_todo(index):
    return {'type': ActionTypes.COMPLETE_TODO, 'index': index}


# initial state
initial_state = {
    'todos': ()
}


# reducers
def todo_app(state, action):
    if action['type'] == ActionTypes.ADD_TODO:
        todos = state['todos'] + (action['text'],)
        return toolz.assoc(state, 'todos', todos)
    elif action['type'] == ActionTypes.COMPLETE_TODO:
        todos = state['todos'][:action['index']] + state['todos'][action['index'] + 1:]
        return toolz.assoc(state, 'todos', todos)
    else:
        return state


@asyncio.coroutine
def run():
    store = aioredux.Store(todo_app, initial_state)
    store.subscribe(lambda: logging.info("new state: {}".format(store.state)))
    for i in range(5):
        store.dispatch(add_todo('do task {}'.format(i)))
    store.dispatch(complete_todo(1))
    store.dispatch(complete_todo(2))
    logging.info('Finished')

if __name__ == '__main__':
    # configure logging
    logging.basicConfig(level=logging.INFO)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(run())



import asyncio


def thunk_middleware(dispatch, state_func):
    def next_func(next_):
        def action_func(action):
            if callable(action):
                if not asyncio.iscoroutinefunction(action):
                    raise RuntimeError('thunk must be a coroutine')
                return action(dispatch, state_func)
            else:
                return next_(action)
        return action_func
    return next_func



import asyncio

import aioredux.utils


class ActionTypes:
    INIT = '@@redux/INIT'


class Store:
    '''Creates a store which holds the state tree.'''

    def __init__(self, reducer, initial_state=None, loop=None):
        '''Do not call directly, use create_store'''
        if not callable(reducer):
            raise ValueError('Expected the reducer to be callable.')
        self.reducer = reducer
        self._state = initial_state
        self.listeners = set()
        self.is_dispatching = False
        self.loop = loop

    @property
    def state(self):
        '''Access current state (read-only).'''
        return self._state

    def subscribe(self, listener):
        '''Add a change listener.

        `listener` is a function which will be called when the state
        tree has changed.
        '''
        self.listeners.add(listener)

        def unsubscribe():
            self.listeners.remove(listener)

        return unsubscribe

    @asyncio.coroutine
    def replace_reducer(self, next_reducer):
        self.reducer = next_reducer
        yield from self.dispatch({'type': ActionTypes.INIT})

    @asyncio.coroutine
    def dispatch(self, action):
        '''Dispatch an action.'''
        if not aioredux.utils.is_mapping(action):
            return ValueError('Actions must be mappings.')
        if self.is_dispatching:
            return RuntimeError('Reducers may not dispatch actions.')
        try:
            self.is_dispatching = True
            next_state = self.reducer(self.state, action)
        finally:
            self.is_dispatching = False
        # If no change in state, do not notify subscribers
        if next_state != self.state:
            self._state = next_state
            for listener in self.listeners:
                listener()
        future = asyncio.Future(loop=self.loop)
        future.set_result(action)
        return future


@asyncio.coroutine
def create_store(*args, **kwargs):
    store = Store(*args, **kwargs)
    # dispatch an 'INIT' action so every reducer returns initial state
    yield from store.dispatch({'type': ActionTypes.INIT})
    return store



import asyncio
import functools

import toolz

import aioredux
import aioredux.middleware
from aioredux.tests import base
import aioredux.utils


class TestThunk(base.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        super().setUp()

    def tearDown(self):
        self.loop.close()
        self.loop = None
        super().tearDown()

    def test_todo_thunk(self):

        @asyncio.coroutine
        def go():
            initial_state = {
                'todos': (),
            }

            def add_todo(text):
                return {'type': 'ADD_TODO', 'text': text}

            def reducer(state, action):
                if action['type'] == 'ADD_TODO':
                    todos = state['todos'] + (action['text'],)
                    return toolz.assoc(state, 'todos', todos)
                return state

            create_store_partial = functools.partial(aioredux.create_store, loop=self.loop)
            create_store_with_middleware = aioredux.apply_middleware(aioredux.middleware.thunk_middleware)(create_store_partial)  # noqa
            store = yield from create_store_with_middleware(reducer, initial_state)
            self.assertIsNotNone(store.state)
            self.assertIsNotNone(store.state['todos'])
            self.assertEqual(len(store.state['todos']), 0)

            @asyncio.coroutine
            def thunk(dispatch, state_func):
                yield from dispatch(add_todo('todo text'))

            yield from store.dispatch(thunk)
            self.assertIsNotNone(store.state)
            self.assertIsNotNone(store.state['todos'])
            self.assertEqual(len(store.state['todos']), 1)

        self.loop.run_until_complete(go())

    def test_todo_thunk_return_value(self):

        @asyncio.coroutine
        def go():
            initial_state = {
                'todos': (),
            }

            def add_todo(text):
                return {'type': 'ADD_TODO', 'text': text}

            def add_todo_complex():
                @asyncio.coroutine
                def thunk(dispatch, state_func):
                    yield from dispatch({'type': 'ADD_TODO', 'text': 'todo text'})
                    return 19
                return thunk

            def reducer(state, action):
                if action['type'] == 'ADD_TODO':
                    todos = state['todos'] + (action['text'],)
                    return toolz.assoc(state, 'todos', todos)
                return state

            create_store_partial = functools.partial(aioredux.create_store, loop=self.loop)
            create_store_with_middleware = aioredux.apply_middleware(aioredux.middleware.thunk_middleware)(create_store_partial)  # noqa
            store = yield from create_store_with_middleware(reducer, initial_state)

            self.assertIsNotNone(store.state)
            self.assertIsNotNone(store.state['todos'])
            self.assertEqual(len(store.state['todos']), 0)
            # use a thunk
            return_value = yield from store.dispatch(add_todo_complex())
            self.assertEqual(return_value, 19)
            self.assertIsNotNone(store.state)
            self.assertIsNotNone(store.state['todos'])
            self.assertEqual(len(store.state['todos']), 1)

        self.loop.run_until_complete(go())

    def test_todo_thunk_async(self):

        @asyncio.coroutine
        def go():
            initial_state = {
                'todos': (),
            }

            def add_todo(text):
                return {'type': 'ADD_TODO', 'text': text}

            def add_todo_slow():
                @asyncio.coroutine
                def thunk(dispatch, state_func):
                    yield from dispatch(add_todo('slow todo'))
                return thunk

            def reducer(state, action):
                if action['type'] == 'ADD_TODO':
                    todos = state['todos'] + (action['text'],)
                    return toolz.assoc(state, 'todos', todos)
                return state

            create_store_partial = functools.partial(aioredux.create_store, loop=self.loop)
            create_store_with_middleware = aioredux.apply_middleware(aioredux.middleware.thunk_middleware)(create_store_partial)  # noqa
            store = yield from create_store_with_middleware(reducer, initial_state)

            self.assertIsNotNone(store.state)
            self.assertIsNotNone(store.state['todos'])
            self.assertEqual(len(store.state['todos']), 0)
            yield from store.dispatch(add_todo('first todo'))
            self.assertIsNotNone(store.state)
            self.assertEqual(len(store.state['todos']), 1)

            yield from store.dispatch(add_todo_slow())
            yield from asyncio.sleep(0.01, loop=self.loop)
            self.assertIsNotNone(store.state)
            self.assertEqual(len(store.state['todos']), 2)

        self.loop.run_until_complete(go())
